# This file is part of CycloneDX Python Library
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.

"""CVSS related utilities"""

__all__ = ['vs_from_cvss_scores']

from typing import Union

from ...model.vulnerability import VulnerabilitySeverity


def vs_from_cvss_scores(scores: Union[tuple[float, ...], float, None]) -> VulnerabilitySeverity:
    """
    Derives the Severity of a Vulnerability from it's declared CVSS scores.

    Args:
        scores: A `tuple` of CVSS scores. CVSS scoring system allows for up to three separate scores.

    Returns:
        Always returns an instance of :class:`cyclonedx.model.vulnerability.VulnerabilitySeverity`.
    """
    if type(scores) is float:
        scores = (scores,)

    if scores is None:
        return VulnerabilitySeverity.UNKNOWN

    max_cvss_score: float
    if isinstance(scores, tuple):
        max_cvss_score = max(scores)
    else:
        max_cvss_score = float(scores)

    if max_cvss_score >= 9.0:
        return VulnerabilitySeverity.CRITICAL
    elif max_cvss_score >= 7.0:
        return VulnerabilitySeverity.HIGH
    elif max_cvss_score >= 4.0:
        return VulnerabilitySeverity.MEDIUM
    elif max_cvss_score > 0.0:
        return VulnerabilitySeverity.LOW
    else:
        return VulnerabilitySeverity.NONE



# This file is part of CycloneDX Python Library
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.


from collections.abc import Generator, Iterable
from datetime import datetime
from enum import Enum
from itertools import chain
from typing import TYPE_CHECKING, Optional, Union
from uuid import UUID, uuid4
from warnings import warn

import py_serializable as serializable
from sortedcontainers import SortedSet

from .._internal.compare import ComparableTuple as _ComparableTuple
from .._internal.time import get_now_utc as _get_now_utc
from ..exception.model import LicenseExpressionAlongWithOthersException, UnknownComponentDependencyException
from ..schema.schema import (
    SchemaVersion1Dot0,
    SchemaVersion1Dot1,
    SchemaVersion1Dot2,
    SchemaVersion1Dot3,
    SchemaVersion1Dot4,
    SchemaVersion1Dot5,
    SchemaVersion1Dot6,
    SchemaVersion1Dot7,
)
from ..serialization import UrnUuidHelper
from . import _BOM_LINK_PREFIX, ExternalReference, Property
from .bom_ref import BomRef
from .component import Component
from .contact import OrganizationalContact, OrganizationalEntity
from .definition import Definitions
from .dependency import Dependable, Dependency
from .license import License, LicenseExpression, LicenseRepository, _LicenseRepositorySerializationHelper
from .lifecycle import Lifecycle, LifecycleRepository, _LifecycleRepositoryHelper
from .service import Service
from .tool import Tool, ToolRepository, _ToolRepositoryHelper
from .vulnerability import Vulnerability

if TYPE_CHECKING:  # pragma: no cover
    from packageurl import PackageURL


@serializable.serializable_enum
class TlpClassification(str, Enum):
    """
    Enum object that defines the Traffic Light Protocol (TLP) classification that controls the sharing and distribution
    of the data that the BOM describes.

    .. note::
        Introduced in CycloneDX v1.7

    .. note::
        See the CycloneDX Schema definition: https://cyclonedx.org/docs/1.7/xml/#type_tlpClassificationType
    """

    CLEAR = 'CLEAR'
    GREEN = 'GREEN'
    AMBER = 'AMBER'
    AMBER_AND_STRICT = 'AMBER_AND_STRICT'
    RED = 'RED'


@serializable.serializable_class(ignore_unknown_during_deserialization=True)
class DistributionConstraints:
    """
    Our internal representation of the `distributionConstraints` complex type.
    Conditions and constraints governing the sharing and distribution of the data or components described by this BOM.

    .. note::
        Introduced in CycloneDX v1.7

    .. note::
        See the CycloneDX Schema definition: https://cyclonedx.org/docs/1.7/xml/#type_metadata
    """

    def __init__(
        self, *,
        tlp: Optional[TlpClassification] = None,
    ) -> None:
        self.tlp = tlp or TlpClassification.CLEAR

    @property
    @serializable.xml_sequence(0)
    def tlp(self) -> TlpClassification:
        """
        The Traffic Light Protocol (TLP) classification that controls the sharing and distribution of the data that the
        BOM describes.

        Returns:
            `TlpClassification` enum value
        """
        return self._tlp

    @tlp.setter
    def tlp(self, tlp: TlpClassification) -> None:
        self._tlp = tlp

    def __comparable_tuple(self) -> _ComparableTuple:
        return _ComparableTuple(self.tlp)

    def __eq__(self, other: object) -> bool:
        if isinstance(other, DistributionConstraints):
            return self.__comparable_tuple() == other.__comparable_tuple()
        return False

    def __lt__(self, other: object) -> bool:
        if isinstance(other, DistributionConstraints):
            return self.__comparable_tuple() < other.__comparable_tuple()
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self.__comparable_tuple())

    def __repr__(self) -> str:
        return f'<DistributionConstraints tlp={self.tlp}>'


@serializable.serializable_class(ignore_unknown_during_deserialization=True)
class BomMetaData:
    """
    This is our internal representation of the metadata complex type within the CycloneDX standard.

    .. note::
        See the CycloneDX Schema for Bom metadata: https://cyclonedx.org/docs/1.7/xml/#type_metadata
    """

    def __init__(
        self, *,
        tools: Optional[Union[Iterable[Tool], ToolRepository]] = None,
        authors: Optional[Iterable[OrganizationalContact]] = None,
        component: Optional[Component] = None,
        supplier: Optional[OrganizationalEntity] = None,
        licenses: Optional[Iterable[License]] = None,
        properties: Optional[Iterable[Property]] = None,
        timestamp: Optional[datetime] = None,
        manufacturer: Optional[OrganizationalEntity] = None,
        lifecycles: Optional[Iterable[Lifecycle]] = None,
        distribution_constraints: Optional[DistributionConstraints] = None,
        # Deprecated as of v1.6
        manufacture: Optional[OrganizationalEntity] = None,
    ) -> None:
        self.timestamp = timestamp or _get_now_utc()
        self.tools = tools or []
        self.authors = authors or []
        self.component = component
        self.supplier = supplier
        self.licenses = licenses or []
        self.properties = properties or []
        self.manufacturer = manufacturer
        self.lifecycles = lifecycles or []
        self.distribution_constraints = distribution_constraints
        # deprecated properties below
        self.manufacture = manufacture

    @property
    @serializable.type_mapping(serializable.helpers.XsdDateTime)
    @serializable.xml_sequence(1)
    def timestamp(self) -> datetime:
        """
        The date and time (in UTC) when this BomMetaData was created.

        Returns:
            `datetime` instance in UTC timezone
        """
        return self._timestamp

    @timestamp.setter
    def timestamp(self, timestamp: datetime) -> None:
        self._timestamp = timestamp

    @property
    @serializable.view(SchemaVersion1Dot5)
    @serializable.view(SchemaVersion1Dot6)
    @serializable.view(SchemaVersion1Dot7)
    @serializable.type_mapping(_LifecycleRepositoryHelper)
    @serializable.xml_sequence(2)
    def lifecycles(self) -> LifecycleRepository:
        """
        An optional list of BOM lifecycle stages.

        Returns:
            Set of `Lifecycle`
        """
        return self._lifecycles

    @lifecycles.setter
    def lifecycles(self, lifecycles: Iterable[Lifecycle]) -> None:
        self._lifecycles = LifecycleRepository(lifecycles)

    @property
    @serializable.type_mapping(_ToolRepositoryHelper)
    @serializable.xml_sequence(3)
    def tools(self) -> ToolRepository:
        """
        Tools used to create this BOM.

        Returns:
            :class:`ToolRepository` object.
        """
        return self._tools

    @tools.setter
    def tools(self, tools: Union[Iterable[Tool], ToolRepository]) -> None:
        self._tools = tools \
            if isinstance(tools, ToolRepository) \
            else ToolRepository(tools=tools)

    @property
    @serializable.xml_array(serializable.XmlArraySerializationType.NESTED, 'author')
    @serializable.xml_sequence(4)
    def authors(self) -> 'SortedSet[OrganizationalContact]':
        """
        The person(s) who created the BOM.

        Authors are common in BOMs created through manual processes.

        BOMs created through automated means may not have authors.

        Returns:
            Set of `OrganizationalContact`
        """
        return self._authors

    @authors.setter
    def authors(self, authors: Iterable[OrganizationalContact]) -> None:
        self._authors = SortedSet(authors)

    @property
    @serializable.xml_sequence(5)
    def component(self) -> Optional[Component]:
        """
        The (optional) component that the BOM describes.

        Returns:
            `cyclonedx.model.component.Component` instance for this Bom Metadata.
        """
        return self._component

    @component.setter
    def component(self, component: Optional[Component]) -> None:
        """
        The (optional) component that the BOM describes.

        Args:
            component
                `cyclonedx.model.component.Component` instance to add to this Bom Metadata.

        Returns:
            None
        """
        self._component = component

    @property
    @serializable.view(SchemaVersion1Dot2)
    @serializable.view(SchemaVersion1Dot3)
    @serializable.view(SchemaVersion1Dot4)
    @serializable.view(SchemaVersion1Dot5)
    @serializable.view(SchemaVersion1Dot6)
    @serializable.view(SchemaVersion1Dot7)
    @serializable.xml_sequence(6)
    def manufacture(self) -> Optional[OrganizationalEntity]:
        """
        The organization that manufactured the component that the BOM describes.

        Returns:
            `OrganizationalEntity` if set else `None`
        """
        return self._manufacture

    @manufacture.setter
    def manufacture(self, manufacture: Optional[OrganizationalEntity]) -> None:
        """
        @todo Based on https://github.com/CycloneDX/specification/issues/346,
              we should set this data on `.component.manufacturer`.
        """
        if manufacture is not None:
            warn(
                '`bom.metadata.manufacture` is deprecated from CycloneDX v1.6 onwards. '
                'Please use `bom.metadata.component.manufacturer` instead.',
                DeprecationWarning)
        self._manufacture = manufacture

    @property
    @serializable.view(SchemaVersion1Dot6)
    @serializable.view(SchemaVersion1Dot7)
    @serializable.xml_sequence(7)
    def manufacturer(self) -> Optional[OrganizationalEntity]:
        """
        The organization that created the BOM.
        Manufacturer is common in BOMs created through automated processes. BOMs created through manual means may have
        `@.authors` instead.

        Returns:
            `OrganizationalEntity` if set else `None`
        """
        return self._manufacturer

    @manufacturer.setter
    def manufacturer(self, manufacturer: Optional[OrganizationalEntity]) -> None:
        self._manufacturer = manufacturer

    @property
    @serializable.xml_sequence(8)
    def supplier(self) -> Optional[OrganizationalEntity]:
        """
        The organization that supplied the component that the BOM describes.

        The supplier may often be the manufacturer, but may also be a distributor or repackager.

        Returns:
            `OrganizationalEntity` if set else `None`
        """
        return self._supplier

    @supplier.setter
    def supplier(self, supplier: Optional[OrganizationalEntity]) -> None:
        self._supplier = supplier

    @property
    @serializable.view(SchemaVersion1Dot3)
    @serializable.view(SchemaVersion1Dot4)
    @serializable.view(SchemaVersion1Dot5)
    @serializable.view(SchemaVersion1Dot6)
    @serializable.view(SchemaVersion1Dot7)
    @serializable.type_mapping(_LicenseRepositorySerializationHelper)
    @serializable.xml_sequence(9)
    def licenses(self) -> LicenseRepository:
        """
        A optional list of statements about how this BOM is licensed.

        Returns:
            Set of `LicenseChoice`
        """
        return self._licenses

    @licenses.setter
    def licenses(self, licenses: Iterable[License]) -> None:
        self._licenses = LicenseRepository(licenses)

    @property
    @serializable.view(SchemaVersion1Dot3)
    @serializable.view(SchemaVersion1Dot4)
    @serializable.view(SchemaVersion1Dot5)
    @serializable.view(SchemaVersion1Dot6)
    @serializable.view(SchemaVersion1Dot7)
    @serializable.xml_array(serializable.XmlArraySerializationType.NESTED, 'property')
    @serializable.xml_sequence(10)
    def properties(self) -> 'SortedSet[Property]':
        """
        Provides the ability to document properties in a key/value store. This provides flexibility to include data not
        officially supported in the standard without having to use additional namespaces or create extensions.

        Property names of interest to the general public are encouraged to be registered in the CycloneDX Property
        Taxonomy - https://github.com/CycloneDX/cyclonedx-property-taxonomy. Formal registration is OPTIONAL.

        Return:
            Set of `Property`
        """
        return self._properties

    @properties.setter
    def properties(self, properties: Iterable[Property]) -> None:
        self._properties = SortedSet(properties)

    @property
    @serializable.view(SchemaVersion1Dot7)
    @serializable.xml_sequence(11)
    def distribution_constraints(self) -> Optional[DistributionConstraints]:
        """
        Conditions and constraints governing the sharing and distribution of the data or components described by this
        BOM.

        Returns:
            `DistributionConstraints` or `None`
        """
        return self._distribution_constraints

    @distribution_constraints.setter
    def distribution_constraints(self, distribution_constraints: Optional[DistributionConstraints]) -> None:
        self._distribution_constraints = distribution_constraints

    def __comparable_tuple(self) -> _ComparableTuple:
        return _ComparableTuple((
            _ComparableTuple(self.authors), self.component, _ComparableTuple(self.licenses), self.manufacture,
            _ComparableTuple(self.properties), self.distribution_constraints,
            _ComparableTuple(self.lifecycles), self.supplier, self.timestamp, self.tools, self.manufacturer
        ))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, BomMetaData):
            return self.__comparable_tuple() == other.__comparable_tuple()
        return False

    def __lt__(self, other: object) -> bool:
        if isinstance(other, BomMetaData):
            return self.__comparable_tuple() == other.__comparable_tuple()
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self.__comparable_tuple())

    def __repr__(self) -> str:
        return f'<BomMetaData timestamp={self.timestamp}, component={self.component}>'


@serializable.serializable_class(
    ignore_during_deserialization={
        '$schema', 'bom_format', 'spec_version',  # JSON-implementation's format hints
    },
    ignore_unknown_during_deserialization=True
)
class Bom:
    """
    This is our internal representation of a bill-of-materials (BOM).

    Once you have an instance of `cyclonedx.model.bom.Bom`, you can pass this to an instance of
    `cyclonedx.output.BaseOutput` to produce a CycloneDX document according to a specific schema version and format.
    """

    def __init__(
        self, *,
        components: Optional[Iterable[Component]] = None,
        services: Optional[Iterable[Service]] = None,
        external_references: Optional[Iterable[ExternalReference]] = None,
        serial_number: Optional[UUID] = None,
        version: int = 1,
        metadata: Optional[BomMetaData] = None,
        dependencies: Optional[Iterable[Dependency]] = None,
        vulnerabilities: Optional[Iterable[Vulnerability]] = None,
        properties: Optional[Iterable[Property]] = None,
        definitions: Optional[Definitions] = None,
    ) -> None:
        """
        Create a new Bom that you can manually/programmatically add data to later.

        Returns:
            New, empty `cyclonedx.model.bom.Bom` instance.
        """
        self.serial_number = serial_number or uuid4()
        self.version = version
        self.metadata = metadata or BomMetaData()
        self.components = components or []
        self.services = services or []
        self.external_references = external_references or []
        self.vulnerabilities = vulnerabilities or []
        self.dependencies = dependencies or []
        self.properties = properties or []
        self.definitions = definitions or Definitions()

    @property
    @serializable.type_mapping(UrnUuidHelper)
    @serializable.view(SchemaVersion1Dot1)
    @serializable.view(SchemaVersion1Dot2)
    @serializable.view(SchemaVersion1Dot3)
    @serializable.view(SchemaVersion1Dot4)
    @serializable.view(SchemaVersion1Dot5)
    @serializable.view(SchemaVersion1Dot6)
    @serializable.view(SchemaVersion1Dot7)
    @serializable.xml_attribute()
    def serial_number(self) -> UUID:
        """
        Unique UUID for this BOM

        Returns:
            `UUID` instance
            `UUID` instance
        """
        return self._serial_number

    @serial_number.setter
    def serial_number(self, serial_number: UUID) -> None:
        self._serial_number = serial_number

    @property
    @serializable.xml_attribute()
    def version(self) -> int:
        return self._version

    @version.setter
    def version(self, version: int) -> None:
        self._version = version

    @property
    @serializable.view(SchemaVersion1Dot2)
    @serializable.view(SchemaVersion1Dot3)
    @serializable.view(SchemaVersion1Dot4)
    @serializable.view(SchemaVersion1Dot5)
    @serializable.view(SchemaVersion1Dot6)
    @serializable.view(SchemaVersion1Dot7)
    @serializable.xml_sequence(10)
    def metadata(self) -> BomMetaData:
        """
        Get our internal metadata object for this Bom.

        Returns:
            Metadata object instance for this Bom.

        .. note::
            See the CycloneDX Schema for Bom metadata: https://cyclonedx.org/docs/1.7/xml/#type_metadata
        """
        return self._metadata

    @metadata.setter
    def metadata(self, metadata: BomMetaData) -> None:
        self._metadata = metadata

    @property
    @serializable.include_none(SchemaVersion1Dot0)
    @serializable.include_none(SchemaVersion1Dot1)
    @serializable.xml_array(serializable.XmlArraySerializationType.NESTED, 'component')
    @serializable.xml_sequence(20)
    def components(self) -> 'SortedSet[Component]':
        """
        Get all the Components currently in this Bom.

        Returns:
             Set of `Component` in this Bom
        """
        return self._components

    @components.setter
    def components(self, components: Iterable[Component]) -> None:
        self._components = SortedSet(components)

    @property
    @serializable.view(SchemaVersion1Dot2)
    @serializable.view(SchemaVersion1Dot3)
    @serializable.view(SchemaVersion1Dot4)
    @serializable.view(SchemaVersion1Dot5)
    @serializable.view(SchemaVersion1Dot6)
    @serializable.view(SchemaVersion1Dot7)
    @serializable.xml_array(serializable.XmlArraySerializationType.NESTED, 'service')
    @serializable.xml_sequence(30)
    def services(self) -> 'SortedSet[Service]':
        """
        Get all the Services currently in this Bom.

        Returns:
             Set of `Service` in this BOM
        """
        return self._services

    @services.setter
    def services(self, services: Iterable[Service]) -> None:
        self._services = SortedSet(services)

    @property
    @serializable.view(SchemaVersion1Dot1)
    @serializable.view(SchemaVersion1Dot2)
    @serializable.view(SchemaVersion1Dot3)
    @serializable.view(SchemaVersion1Dot4)
    @serializable.view(SchemaVersion1Dot5)
    @serializable.view(SchemaVersion1Dot6)
    @serializable.view(SchemaVersion1Dot7)
    @serializable.xml_array(serializable.XmlArraySerializationType.NESTED, 'reference')
    @serializable.xml_sequence(40)
    def external_references(self) -> 'SortedSet[ExternalReference]':
        """
        Provides the ability to document external references related to the BOM or to the project the BOM describes.

        Returns:
            Set of `ExternalReference`
        """
        return self._external_references

    @external_references.setter
    def external_references(self, external_references: Iterable[ExternalReference]) -> None:
        self._external_references = SortedSet(external_references)

    @property
    @serializable.view(SchemaVersion1Dot2)
    @serializable.view(SchemaVersion1Dot3)
    @serializable.view(SchemaVersion1Dot4)
    @serializable.view(SchemaVersion1Dot5)
    @serializable.view(SchemaVersion1Dot6)
    @serializable.view(SchemaVersion1Dot7)
    @serializable.xml_array(serializable.XmlArraySerializationType.NESTED, 'dependency')
    @serializable.xml_sequence(50)
    def dependencies(self) -> 'SortedSet[Dependency]':
        return self._dependencies

    @dependencies.setter
    def dependencies(self, dependencies: Iterable[Dependency]) -> None:
        self._dependencies = SortedSet(dependencies)

    # @property
    # ...
    # @serializable.view(SchemaVersion1Dot3)
    # @serializable.view(SchemaVersion1Dot4)
    # @serializable.view(SchemaVersion1Dot5)
    # @serializable.xml_sequence(6)
    # def compositions(self) -> ...:
    #     ...  # TODO Since CDX 1.3
    #
    # @compositions.setter
    # def compositions(self, ...) -> None:
    #     ...  # TODO Since CDX 1.3

    @property
    # @serializable.view(SchemaVersion1Dot3) @todo: Update py-serializable to support view by OutputFormat filtering
    # @serializable.view(SchemaVersion1Dot4) @todo: Update py-serializable to support view by OutputFormat filtering
    @serializable.view(SchemaVersion1Dot5)
    @serializable.view(SchemaVersion1Dot6)
    @serializable.view(SchemaVersion1Dot7)
    @serializable.xml_array(serializable.XmlArraySerializationType.NESTED, 'property')
    @serializable.xml_sequence(70)
    def properties(self) -> 'SortedSet[Property]':
        """
        Provides the ability to document properties in a name/value store. This provides flexibility to include data
        not officially supported in the standard without having to use additional namespaces or create extensions.
        Property names of interest to the general public are encouraged to be registered in the CycloneDX Property
        Taxonomy - https://github.com/CycloneDX/cyclonedx-property-taxonomy. Formal registration is OPTIONAL.

        Return:
            Set of `Property`
        """
        return self._properties

    @properties.setter
    def properties(self, properties: Iterable[Property]) -> None:
        self._properties = SortedSet(properties)

    @property
    @serializable.view(SchemaVersion1Dot4)
    @serializable.view(SchemaVersion1Dot5)
    @serializable.view(SchemaVersion1Dot6)
    @serializable.view(SchemaVersion1Dot7)
    @serializable.xml_array(serializable.XmlArraySerializationType.NESTED, 'vulnerability')
    @serializable.xml_sequence(80)
    def vulnerabilities(self) -> 'SortedSet[Vulnerability]':
        """
        Get all the Vulnerabilities in this BOM.

        Returns:
             Set of `Vulnerability`
        """
        return self._vulnerabilities

    @vulnerabilities.setter
    def vulnerabilities(self, vulnerabilities: Iterable[Vulnerability]) -> None:
        self._vulnerabilities = SortedSet(vulnerabilities)

    # @property
    # ...
    # @serializable.view(SchemaVersion1Dot5)
    # @serializable.xml_sequence(9)
    # def annotations(self) -> ...:
    #     ... # TODO Since CDX 1.5
    #
    # @annotations.setter
    # def annotations(self, ...) -> None:
    #     ...  # TODO Since CDX 1.5

    # @property
    # ...
    # @serializable.view(SchemaVersion1Dot5)
    # @formulation.xml_sequence(10)
    # def formulation(self) -> ...:
    #     ... # TODO Since CDX 1.5
    #
    # @formulation.setter
    # def formulation(self, ...) -> None:
    #     ...  # TODO Since CDX 1.5

    @property
    @serializable.view(SchemaVersion1Dot6)
    @serializable.view(SchemaVersion1Dot7)
    @serializable.xml_sequence(110)
    def definitions(self) -> Optional[Definitions]:
        """
        The repository for definitions

        Returns:
            `Definitions`
        """
        return self._definitions if len(self._definitions.standards) > 0 else None

    @definitions.setter
    def definitions(self, definitions: Definitions) -> None:
        self._definitions = definitions

    def get_component_by_purl(self, purl: Optional['PackageURL']) -> Optional[Component]:
        """
        Get a Component already in the Bom by its PURL

        Args:
             purl:
                An instance of `packageurl.PackageURL` to look and find `Component`.

        Returns:
            `Component` or `None`

        .. deprecated:: next
        """
        if purl:
            found = [x for x in self.components if x.purl == purl]
            if len(found) == 1:
                return found[0]

        return None

    def get_urn_uuid(self) -> str:
        """
        Get the unique reference for this Bom.

        Returns:
            URN formatted UUID that uniquely identified this Bom instance.

        .. deprecated:: next
        """
        return self.serial_number.urn

    def has_component(self, component: Component) -> bool:
        """
        Check whether this Bom contains the provided Component.

        Args:
            component:
                The instance of `cyclonedx.model.component.Component` to check if this Bom contains.

        Returns:
            `bool` - `True` if the supplied Component is part of this Bom, `False` otherwise.

        .. deprecated:: next
        """
        return component in self.components

    def _get_all_components(self) -> Generator[Component, None, None]:
        if self.metadata.component:
            yield from self.metadata.component.get_all_nested_components(include_self=True)
        for c in self.components:
            yield from c.get_all_nested_components(include_self=True)

    def get_vulnerabilities_for_bom_ref(self, bom_ref: BomRef) -> 'SortedSet[Vulnerability]':
        """
        Get all known Vulnerabilities that affect the supplied bom_ref.

        Args:
            bom_ref: `BomRef`

        Returns:
            `SortedSet` of `Vulnerability`

        .. deprecated:: next
            Deprecated without any replacement.
        """
        vulnerabilities: SortedSet[Vulnerability] = SortedSet()
        for v in self.vulnerabilities:
            for target in v.affects:
                if target.ref == bom_ref.value:
                    vulnerabilities.add(v)
        return vulnerabilities

    def has_vulnerabilities(self) -> bool:
        """
        Check whether this Bom has any declared vulnerabilities.

        Returns:
            `bool` - `True` if this Bom has at least one Vulnerability, `False` otherwise.

        .. deprecated:: next
            Deprecated without any replacement.
        """
        return bool(self.vulnerabilities)

    def register_dependency(self, target: Dependable, depends_on: Optional[Iterable[Dependable]] = None) -> None:
        _d = next(filter(lambda _d: _d.ref == target.bom_ref, self.dependencies), None)
        if _d:
            # Dependency Target already registered - but it might have new dependencies to add
            if depends_on:
                _d.dependencies.update(map(lambda _d: Dependency(ref=_d.bom_ref), depends_on))
        else:
            # First time we are seeing this target as a Dependency
            self._dependencies.add(Dependency(
                ref=target.bom_ref,
                dependencies=map(lambda _dep: Dependency(ref=_dep.bom_ref), depends_on) if depends_on else []
            ))

        if depends_on:
            # Ensure dependents are registered with no further dependents in the DependencyGraph
            for _d2 in depends_on:
                self.register_dependency(target=_d2, depends_on=None)

    def urn(self) -> str:
        """
        .. deprecated:: next
            Deprecated without any replacement.
        """
        # idea: have 'serial_number' be a string, and use it instead of this method
        return f'{_BOM_LINK_PREFIX}{self.serial_number}/{self.version}'

    def validate(self) -> bool:
        """
        Perform data-model level validations to make sure we have some known data integrity prior to attempting output
        of this `Bom`

        Returns:
             `bool`

        .. deprecated:: next
            Deprecated without any replacement.
        """
        # !! deprecated function. have this as an part of the normalization process, like the BomRefDiscrimator
        # 0. Make sure all Dependable have a Dependency entry
        if self.metadata.component:
            self.register_dependency(target=self.metadata.component)
        for _c in self.components:
            self.register_dependency(target=_c)
        for _s in self.services:
            self.register_dependency(target=_s)

        # 1. Make sure dependencies are all in this Bom.
        component_bom_refs = set(map(lambda c: c.bom_ref, self._get_all_components())) | set(
            map(lambda s: s.bom_ref, self.services))
        dependency_bom_refs = set(chain(
            (d.ref for d in self.dependencies),
            chain.from_iterable(d.dependencies_as_bom_refs() for d in self.dependencies)
        ))
        dependency_diff = dependency_bom_refs - component_bom_refs
        if len(dependency_diff) > 0:
            raise UnknownComponentDependencyException(
                'One or more Components have Dependency references to Components/Services that are not known in this '
                f'BOM. They are: {dependency_diff}')

        # 2. if root component is set and there are other components: dependencies should exist for the Component
        # this BOM is describing
        if self.metadata.component and len(self.components) > 0 and not any(map(
            lambda d: d.ref == self.metadata.component.bom_ref and len(d.dependencies) > 0,  # type:ignore[union-attr]
            self.dependencies
        )):
            warn(
                f'The Component this BOM is describing {self.metadata.component.purl} has no defined dependencies '
                'which means the Dependency Graph is incomplete - you should add direct dependencies to this '
                '"root" Component to complete the Dependency Graph data.',
                category=UserWarning, stacklevel=1
            )

        # 3. If a LicenseExpression is set, then there must be no other license.
        # see https://github.com/CycloneDX/specification/pull/205
        elem: Union[BomMetaData, Component, Service]
        for elem in chain(  # type:ignore[assignment]
            [self.metadata],
            self.metadata.component.get_all_nested_components(include_self=True) if self.metadata.component else [],
            chain.from_iterable(c.get_all_nested_components(include_self=True) for c in self.components),
            self.services
        ):
            if len(elem.licenses) > 1 and any(isinstance(li, LicenseExpression) for li in elem.licenses):
                raise LicenseExpressionAlongWithOthersException(
                    f'Found LicenseExpression along with others licenses in: {elem!r}')

        return True

    def __comparable_tuple(self) -> _ComparableTuple:
        return _ComparableTuple((
            self.serial_number, self.version, self.metadata, _ComparableTuple(
                self.components), _ComparableTuple(self.services),
            _ComparableTuple(self.external_references), _ComparableTuple(
                self.dependencies), _ComparableTuple(self.properties),
            _ComparableTuple(self.vulnerabilities),
        ))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Bom):
            return self.__comparable_tuple() == other.__comparable_tuple()
        return False

    def __hash__(self) -> int:
        return hash(self.__comparable_tuple())

    def __repr__(self) -> str:
        return f'<Bom uuid={self.serial_number}, hash={hash(self)}>'



# This file is part of CycloneDX Python Library
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.


"""
Set of classes and methods for outputting our libraries internal Bom model to CycloneDX documents in varying formats
and according to different versions of the CycloneDX schema standard.
"""

import os
from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping
from itertools import chain
from random import random
from typing import TYPE_CHECKING, Any, Literal, Optional, Union, overload

from ..schema import OutputFormat, SchemaVersion

if TYPE_CHECKING:  # pragma: no cover
    from ..model.bom import Bom
    from ..model.bom_ref import BomRef
    from .json import Json as JsonOutputter
    from .xml import Xml as XmlOutputter


class BaseOutput(ABC):

    def __init__(self, bom: 'Bom', **kwargs: int) -> None:
        super().__init__(**kwargs)
        self._bom = bom
        self._generated: bool = False

    @property
    @abstractmethod
    def schema_version(self) -> SchemaVersion:
        ...  # pragma: no cover

    @property
    @abstractmethod
    def output_format(self) -> OutputFormat:
        ...  # pragma: no cover

    @property
    def generated(self) -> bool:
        return self._generated

    @generated.setter
    def generated(self, generated: bool) -> None:
        self._generated = generated

    def get_bom(self) -> 'Bom':
        return self._bom

    def set_bom(self, bom: 'Bom') -> None:
        self._bom = bom

    @abstractmethod
    def generate(self, force_regeneration: bool = False) -> None:
        ...  # pragma: no cover

    @abstractmethod
    def output_as_string(self, *,
                         indent: Optional[Union[int, str]] = None,
                         **kwargs: Any) -> str:
        ...  # pragma: no cover

    def output_to_file(self, filename: str, allow_overwrite: bool = False, *,
                       indent: Optional[Union[int, str]] = None,
                       **kwargs: Any) -> None:
        # Check directory writable
        output_filename = os.path.realpath(filename)
        output_directory = os.path.dirname(output_filename)
        if not os.access(output_directory, os.W_OK):
            raise PermissionError(output_directory)
        if os.path.exists(output_filename) and not allow_overwrite:
            raise FileExistsError(output_filename)
        with open(output_filename, mode='wb') as f_out:
            f_out.write(self.output_as_string(indent=indent).encode('utf-8'))


@overload
def make_outputter(bom: 'Bom', output_format: Literal[OutputFormat.JSON],
                   schema_version: SchemaVersion) -> 'JsonOutputter':
    ...  # pragma: no cover


@overload
def make_outputter(bom: 'Bom', output_format: Literal[OutputFormat.XML],
                   schema_version: SchemaVersion) -> 'XmlOutputter':
    ...  # pragma: no cover


@overload
def make_outputter(bom: 'Bom', output_format: OutputFormat,
                   schema_version: SchemaVersion) -> Union['XmlOutputter', 'JsonOutputter']:
    ...  # pragma: no cover


def make_outputter(bom: 'Bom', output_format: OutputFormat, schema_version: SchemaVersion) -> BaseOutput:
    """
    Helper method to quickly get the correct output class/formatter.

    Pass in your BOM and optionally an output format and schema version (defaults to XML and latest schema version).


    Raises error when no instance could be made.

    :param bom: Bom
    :param output_format: OutputFormat
    :param schema_version: SchemaVersion
    :return: BaseOutput
    """
    if TYPE_CHECKING:  # pragma: no cover
        BY_SCHEMA_VERSION: Mapping[SchemaVersion, type[BaseOutput]]  # noqa:N806
    if OutputFormat.JSON is output_format:
        from .json import BY_SCHEMA_VERSION
    elif OutputFormat.XML is output_format:
        from .xml import BY_SCHEMA_VERSION
    else:
        raise ValueError(f'Unexpected output_format: {output_format!r}')

    klass = BY_SCHEMA_VERSION.get(schema_version, None)
    if klass is None:
        raise ValueError(f'Unknown {output_format.name}/schema_version: {schema_version!r}')
    return klass(bom)


class BomRefDiscriminator:

    def __init__(self, bomrefs: Iterable['BomRef'], prefix: str = 'BomRef') -> None:
        # do not use dict/set here, different BomRefs with same value have same hash and would shadow each other
        self._bomrefs = tuple((bomref, bomref.value) for bomref in bomrefs)
        self._prefix = prefix

    def __enter__(self) -> None:
        self.discriminate()

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        self.reset()

    def discriminate(self) -> None:
        known_values = []
        for bomref, _ in self._bomrefs:
            value = bomref.value
            if value is None or value in known_values:
                value = self._make_unique()
                bomref.value = value
            known_values.append(value)

    def reset(self) -> None:
        for bomref, original_value in self._bomrefs:
            bomref.value = original_value

    def _make_unique(self) -> str:
        return f'{self._prefix}{str(random())[1:]}{str(random())[1:]}'  # nosec B311

    @classmethod
    def from_bom(cls, bom: 'Bom', prefix: str = 'BomRef') -> 'BomRefDiscriminator':
        return cls(chain(
            map(lambda c: c.bom_ref, bom._get_all_components()),
            map(lambda s: s.bom_ref, bom.services),
            map(lambda v: v.bom_ref, bom.vulnerabilities)
        ), prefix)



# This file is part of CycloneDX Python Library
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.


from unittest import TestCase

from ddt import ddt, named_data

from cyclonedx.exception.model import InvalidCreIdException
from cyclonedx.model.definition import CreId, Definitions, Level, Requirement, Standard


class TestModelDefinitions(TestCase):

    def test_init(self) -> Definitions:
        s = Standard(name='test-standard')
        dr = Definitions(
            standards=(s, ),
        )
        self.assertEqual(1, len(dr.standards))
        self.assertIs(s, tuple(dr.standards)[0])
        return dr

    def test_filled(self) -> None:
        dr = self.test_init()
        self.assertIsNotNone(dr.standards)
        self.assertTrue(dr)

    def test_empty(self) -> None:
        dr = Definitions()
        self.assertIsNotNone(dr.standards)
        self.assertEqual(0, len(dr.standards))
        self.assertFalse(dr)

    def test_unequal_different_type(self) -> None:
        dr = Definitions()
        self.assertFalse(dr == 'other')

    def test_equal_self(self) -> None:
        dr = Definitions()
        dr.standards.add(Standard(name='my-standard'))
        self.assertTrue(dr == dr)

    def test_unequal(self) -> None:
        dr1 = Definitions()
        dr1.standards.add(Standard(name='my-standard'))
        tr2 = Definitions()
        self.assertFalse(dr1 == tr2)

    def test_equal(self) -> None:
        s = Standard(name='my-standard')
        dr1 = Definitions()
        dr1.standards.add(s)
        tr2 = Definitions()
        tr2.standards.add(s)
        self.assertTrue(dr1 == tr2)


@ddt
class TestModelCreId(TestCase):

    def test_different(self) -> None:
        id1 = CreId('CRE:123-456')
        id2 = CreId('CRE:987-654')
        self.assertNotEqual(id(id1), id(id2))
        self.assertNotEqual(hash(id1), hash(id2))
        self.assertFalse(id1 == id2)

    def test_same(self) -> None:
        id1 = CreId('CRE:123-456')
        id2 = CreId('CRE:123-456')
        self.assertNotEqual(id(id1), id(id2))
        self.assertEqual(hash(id1), hash(id2))
        self.assertTrue(id1 == id2)

    def test_invalid_no_id(self) -> None:
        with self.assertRaises(TypeError):
            CreId()

    @named_data(
        ['empty', ''],
        ['arbitrary string', 'some string'],
        ['missing prefix', '123-456'],
        ['additional part', 'CRE:123-456-789'],
        ['no numbers', 'CRE:abc-def'],
        ['no delimiter', 'CRE:123456'],
    )
    def test_invalid_id(self, wrong_id: str) -> None:
        with self.assertRaises(InvalidCreIdException):
            CreId(wrong_id)


class TestModelRequirements(TestCase):

    def test_bom_ref_is_set_from_value(self) -> None:
        r = Requirement(bom_ref='123-456')
        self.assertIsNotNone(r.bom_ref)
        self.assertEqual('123-456', r.bom_ref.value)

    def test_bom_ref_is_set_if_none_given(self) -> None:
        r = Requirement()
        self.assertIsNotNone(r.bom_ref)


class TestModelLevel(TestCase):

    def test_bom_ref_is_set_from_value(self) -> None:
        r = Level(bom_ref='123-456')
        self.assertIsNotNone(r.bom_ref)
        self.assertEqual('123-456', r.bom_ref.value)

    def test_bom_ref_is_set_if_none_given(self) -> None:
        r = Level()
        self.assertIsNotNone(r.bom_ref)


class TestModelStandard(TestCase):

    def test_bom_ref_is_set_from_value(self) -> None:
        r = Standard(bom_ref='123-456')
        self.assertIsNotNone(r.bom_ref)
        self.assertEqual('123-456', r.bom_ref.value)

    def test_bom_ref_is_set_if_none_given(self) -> None:
        r = Standard()
        self.assertIsNotNone(r.bom_ref)



# This file is part of CycloneDX Python Library
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.


import re
from collections.abc import Callable
from typing import Any
from unittest import TestCase
from unittest.mock import Mock, patch
from warnings import warn

from ddt import data, ddt, idata, named_data, unpack

from cyclonedx.exception import CycloneDxException, MissingOptionalDependencyException
from cyclonedx.exception.model import (
    InvalidOmniBorIdException,
    InvalidSwhidException,
    LicenseExpressionAlongWithOthersException,
    UnknownComponentDependencyException,
)
from cyclonedx.exception.output import FormatNotSupportedException
from cyclonedx.model.bom import Bom
from cyclonedx.output.json import BY_SCHEMA_VERSION, Json
from cyclonedx.schema import OutputFormat, SchemaVersion
from cyclonedx.validation.json import JsonStrictValidator
from tests import SnapshotMixin, is_valid_for_schema_version, mksname
from tests._data.models import all_get_bom_funct_invalid, all_get_bom_funct_valid, bom_all_same_bomref

UNSUPPORTED_SV = frozenset((SchemaVersion.V1_1, SchemaVersion.V1_0,))


@ddt
class TestOutputJson(TestCase, SnapshotMixin):

    @data(*UNSUPPORTED_SV)
    def test_unsupported_schema_raises(self, sv: SchemaVersion) -> None:
        outputter_class = BY_SCHEMA_VERSION[sv]
        self.assertTrue(issubclass(outputter_class, Json))
        outputter = outputter_class(Mock(spec=Bom))
        with self.assertRaises(FormatNotSupportedException):
            outputter.output_as_string()

    @named_data(*(
        (f'{n}-{sv.to_version()}', gb, sv)
        for n, gb in all_get_bom_funct_valid
        for sv in SchemaVersion
        if sv not in UNSUPPORTED_SV
        and is_valid_for_schema_version(gb, sv)
    ))
    @unpack
    @patch('cyclonedx.contrib.this.builders.__ThisVersion', 'TESTING')
    def test_valid(self, get_bom: Callable[[], Bom], sv: SchemaVersion, *_: Any, **__: Any) -> None:
        snapshot_name = mksname(get_bom, sv, OutputFormat.JSON)
        bom = get_bom()
        json = BY_SCHEMA_VERSION[sv](bom).output_as_string(indent=2)
        try:
            errors = JsonStrictValidator(sv).validate_str(json)
        except MissingOptionalDependencyException:
            warn('!!! skipped schema validation',
                 category=UserWarning, stacklevel=0)
        else:
            self.assertIsNone(errors, json)
        self.assertEqualSnapshot(json, snapshot_name)

    @named_data(*(
        (f'{n}-{sv.to_version()}', gb, sv)
        for n, gb in all_get_bom_funct_invalid
        for sv in SchemaVersion
        if sv not in UNSUPPORTED_SV
        and is_valid_for_schema_version(gb, sv)
    ))
    @unpack
    def test_invalid(self, get_bom: Callable[[], Bom], sv: SchemaVersion) -> None:
        with self.assertRaises(CycloneDxException) as error:
            bom = get_bom()
            outputter = BY_SCHEMA_VERSION[sv](bom)
            outputter.output_as_string()
        if isinstance(error.exception, (
            LicenseExpressionAlongWithOthersException,
            InvalidOmniBorIdException,
            InvalidSwhidException,
            UnknownComponentDependencyException,
        )):
            return None  # expected
        raise error.exception

    def test_bomref_not_duplicate(self) -> None:
        bom, nr_bomrefs = bom_all_same_bomref()
        output = BY_SCHEMA_VERSION[SchemaVersion.V1_4](bom).output_as_string()
        found = re.findall(r'"bom-ref":\s*"(.*?)"', output)
        self.assertEqual(nr_bomrefs, len(found))
        self.assertCountEqual(set(found), found, 'expected unique items')


@ddt
class TestFunctionalBySchemaVersion(TestCase):

    @idata(SchemaVersion)
    def test_get_outputter_expected(self, sv: SchemaVersion) -> None:
        outputter_class = BY_SCHEMA_VERSION[sv]
        self.assertTrue(issubclass(outputter_class, Json))
        outputter = outputter_class(Mock(spec=Bom))
        self.assertIs(outputter.schema_version, sv)
        self.assertIs(outputter.output_format, OutputFormat.JSON)



# This file is part of CycloneDX Python Library
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.

import sys
from typing import TYPE_CHECKING

from packageurl import PackageURL

from cyclonedx.contrib.license.factories import LicenseFactory
from cyclonedx.contrib.this.builders import this_component as cdx_lib_component
from cyclonedx.exception import MissingOptionalDependencyException
from cyclonedx.model import XsUri
from cyclonedx.model.bom import Bom
from cyclonedx.model.component import Component, ComponentType
from cyclonedx.model.contact import OrganizationalEntity
from cyclonedx.output import make_outputter
from cyclonedx.output.json import JsonV1Dot5
from cyclonedx.schema import OutputFormat, SchemaVersion
from cyclonedx.validation import make_schemabased_validator
from cyclonedx.validation.json import JsonStrictValidator

if TYPE_CHECKING:
    from cyclonedx.output.json import Json as JsonOutputter
    from cyclonedx.output.xml import Xml as XmlOutputter
    from cyclonedx.validation.xml import XmlValidator


lc_factory = LicenseFactory()

# region build the BOM

bom = Bom()
bom.metadata.tools.components.add(cdx_lib_component())
bom.metadata.tools.components.add(Component(
    name='my-own-SBOM-generator',
    type=ComponentType.APPLICATION,
))

bom.metadata.component = root_component = Component(
    name='myApp',
    type=ComponentType.APPLICATION,
    licenses=[lc_factory.make_from_string('MIT')],
    bom_ref='myApp',
)

component1 = Component(
    type=ComponentType.LIBRARY,
    name='some-component',
    group='acme',
    version='1.33.7-beta.1',
    licenses=[lc_factory.make_from_string('(c) 2021 Acme inc.')],
    supplier=OrganizationalEntity(
        name='Acme Inc',
        urls=[XsUri('https://www.acme.org')]
    ),
    bom_ref='myComponent@1.33.7-beta.1',
    purl=PackageURL('generic', 'acme', 'some-component', '1.33.7-beta.1')
)
bom.components.add(component1)
bom.register_dependency(root_component, [component1])

component2 = Component(
    type=ComponentType.LIBRARY,
    name='some-library',
    licenses=[lc_factory.make_from_string('GPL-3.0-only WITH Classpath-exception-2.0')]
)
bom.components.add(component2)
bom.register_dependency(component1, [component2])

# endregion build the BOM

# region JSON
"""demo with explicit instructions for SchemaVersion, outputter and validator"""

my_json_outputter: 'JsonOutputter' = JsonV1Dot5(bom)
serialized_json = my_json_outputter.output_as_string(indent=2)
print(serialized_json)
my_json_validator = JsonStrictValidator(SchemaVersion.V1_7)
try:
    json_validation_errors = my_json_validator.validate_str(serialized_json)
    if json_validation_errors:
        print('JSON invalid', 'ValidationError:', repr(json_validation_errors), sep='\n', file=sys.stderr)
        sys.exit(2)
    print('JSON valid')
except MissingOptionalDependencyException as error:
    print('JSON-validation was skipped due to', error)

# endregion JSON

print('', '=' * 30, '', sep='\n')

# region XML
"""demo with implicit instructions for SchemaVersion, outputter and validator. TypeCheckers will catch errors."""

my_xml_outputter: 'XmlOutputter' = make_outputter(bom, OutputFormat.XML, SchemaVersion.V1_7)
serialized_xml = my_xml_outputter.output_as_string(indent=2)
print(serialized_xml)
my_xml_validator: 'XmlValidator' = make_schemabased_validator(
    my_xml_outputter.output_format, my_xml_outputter.schema_version)
try:
    xml_validation_errors = my_xml_validator.validate_str(serialized_xml)
    if xml_validation_errors:
        print('XML invalid', 'ValidationError:', repr(xml_validation_errors), sep='\n', file=sys.stderr)
        sys.exit(2)
    print('XML valid')
except MissingOptionalDependencyException as error:
    print('XML-validation was skipped due to', error)

# endregion XML



# This file is part of CycloneDX Python Library
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.


from random import shuffle
from unittest import TestCase
from unittest.mock import MagicMock

from cyclonedx.exception.model import MutuallyExclusivePropertiesException
from cyclonedx.model import AttachedText, XsUri
from cyclonedx.model.license import DisjunctiveLicense, LicenseExpression
from tests import reorder


class TestModelDisjunctiveLicense(TestCase):
    def test_create_complete_id(self) -> None:
        text = MagicMock(spec=AttachedText)
        url = MagicMock(spec=XsUri)
        license = DisjunctiveLicense(id='foo', text=text, url=url)
        self.assertEqual('foo', license.id)
        self.assertIsNone(license.name)
        self.assertIs(text, license.text)
        self.assertIs(url, license.url)

    def test_update_id_name(self) -> None:
        license = DisjunctiveLicense(id='foo')
        self.assertEqual('foo', license.id)
        self.assertIsNone(license.name)
        license.name = 'bar'
        self.assertIsNone(license.id)
        self.assertEqual('bar', license.name)

    def test_create_complete_named(self) -> None:
        text = MagicMock(spec=AttachedText)
        url = MagicMock(spec=XsUri)
        license = DisjunctiveLicense(name='foo', text=text, url=url)
        self.assertIsNone(license.id)
        self.assertEqual('foo', license.name)
        self.assertIs(text, license.text)
        self.assertIs(url, license.url)

    def test_update_name_id(self) -> None:
        license = DisjunctiveLicense(name='foo')
        self.assertEqual('foo', license.name)
        self.assertIsNone(license.id)
        license.id = 'bar'
        self.assertIsNone(license.name)
        self.assertEqual('bar', license.id)

    def test_throws_when_no_id_nor_name(self) -> None:
        with self.assertRaises(MutuallyExclusivePropertiesException):
            DisjunctiveLicense(id=None, name=None)

    def test_prefers_id_over_name(self) -> None:
        with self.assertWarnsRegex(
                RuntimeWarning,
                'Both `id` and `name` have been supplied - `name` will be ignored!'):
            license = DisjunctiveLicense(id='foo', name='bar')
        self.assertEqual('foo', license.id)
        self.assertEqual(None, license.name)

    def test_equal(self) -> None:
        a = DisjunctiveLicense(id='foo', name='bar')
        b = DisjunctiveLicense(id='foo', name='bar')
        c = DisjunctiveLicense(id='bar', name='foo')
        self.assertEqual(a, b)
        self.assertNotEqual(a, c)
        self.assertNotEqual(a, 'foo')


class TestModelLicenseExpression(TestCase):
    def test_create(self) -> None:
        license = LicenseExpression('foo')
        self.assertEqual('foo', license.value)

    def test_update(self) -> None:
        license = LicenseExpression('foo')
        self.assertEqual('foo', license.value)
        license.value = 'bar'
        self.assertEqual('bar', license.value)

    def test_equal(self) -> None:
        a = LicenseExpression('foo')
        b = LicenseExpression('foo')
        c = LicenseExpression('bar')
        self.assertEqual(a, b)
        self.assertNotEqual(a, c)
        self.assertNotEqual(a, 'foo')


class TestModelLicense(TestCase):

    def test_sort_mixed(self) -> None:
        expected_order = [1, 2, 0]
        licenses = [
            DisjunctiveLicense(name='my license'),
            LicenseExpression(value='MIT or Apache-2.0'),
            DisjunctiveLicense(id='MIT'),
        ]
        expected_licenses = reorder(licenses, expected_order)
        shuffle(licenses)
        sorted_licenses = sorted(licenses)
        self.assertListEqual(sorted_licenses, expected_licenses)



# This file is part of CycloneDX Python Library
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.

from collections.abc import Callable
from os.path import join
from typing import Any
from unittest import TestCase
from unittest.mock import patch

from ddt import ddt, named_data

from cyclonedx.model.bom import Bom
from cyclonedx.schema import OutputFormat, SchemaVersion
from tests import OWN_DATA_DIRECTORY, DeepCompareMixin, SnapshotMixin, mksname
from tests._data.models import (
    all_get_bom_funct_valid_immut,
    all_get_bom_funct_valid_reversible_migrate,
    all_get_bom_funct_with_incomplete_deps,
)

_LATEST_SCHEMA = SchemaVersion.V1_7


@ddt
class TestDeserializeXml(TestCase, SnapshotMixin, DeepCompareMixin):

    @named_data(*all_get_bom_funct_valid_immut,
                *all_get_bom_funct_valid_reversible_migrate)
    @patch('cyclonedx.contrib.this.builders.__ThisVersion', 'TESTING')
    def test_prepared(self, get_bom: Callable[[], Bom], *_: Any, **__: Any) -> None:
        # only latest schema will have all data populated in serialized form
        snapshot_name = mksname(get_bom, _LATEST_SCHEMA, OutputFormat.XML)
        expected = get_bom()
        with open(self.getSnapshotFile(snapshot_name)) as s:
            bom = Bom.from_xml(s)
        self.assertBomDeepEqual(expected, bom,
                                fuzzy_deps=get_bom in all_get_bom_funct_with_incomplete_deps)

    def test_component_evidence_identity(self) -> None:
        xml_file = join(OWN_DATA_DIRECTORY, 'xml',
                        SchemaVersion.V1_6.to_version(),
                        'component_evidence_identity.xml')
        with open(xml_file) as f:
            bom: Bom = Bom.from_xml(f)  # <<< is expected to not crash
        self.assertIsNotNone(bom)



# This file is part of CycloneDX Python Library
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.


from unittest import TestCase

from cyclonedx.model.tool import Tool
from tests import reorder


class TestModelTool(TestCase):

    def test_sort(self) -> None:
        # expected sort order: (vendor, name, version)
        expected_order = [0, 1, 2, 3, 4, 5, 6]
        tools = [
            Tool(vendor='a', name='a', version='1.0.0'),
            Tool(vendor='a', name='a', version='2.0.0'),
            Tool(vendor='a', name='b', version='1.0.0'),
            Tool(vendor='a', name='b'),
            Tool(vendor='b', name='a'),
            Tool(vendor='b', name='b', version='1.0.0'),
            Tool(name='b'),
        ]
        sorted_tools = sorted(tools)
        expected_tools = reorder(tools, expected_order)
        self.assertListEqual(sorted_tools, expected_tools)

    def test_non_equal_tool_and_invalid(self) -> None:
        t = Tool(vendor='VendorA')
        self.assertFalse(t == 'INVALID')

    def test_invalid_tool_compare(self) -> None:
        t = Tool(vendor='VendorA')
        with self.assertRaises(TypeError):
            r = t < 'INVALID'  # pylint: disable=unused-variable # noqa: disable=E841

    def test_tool_repr(self) -> None:
        t = Tool(name='test-tool', version='1.2.3', vendor='test-vendor')
        self.assertEqual(repr(t), '<Tool name=test-tool, version=1.2.3, vendor=test-vendor>')

    def test_tool_equals(self) -> None:
        t = Tool()
        self.assertEqual(t, t)



# This file is part of CycloneDX Python Library
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.


"""
This set of classes represents the data that is possible about known Services.

.. note::
    See the CycloneDX Schema extension definition https://cyclonedx.org/docs/1.7/xml/#type_servicesType
"""


from collections.abc import Iterable
from typing import Any, Optional, Union

import py_serializable as serializable
from sortedcontainers import SortedSet

from .._internal.bom_ref import bom_ref_from_str as _bom_ref_from_str
from .._internal.compare import ComparableTuple as _ComparableTuple
from ..schema.schema import (
    SchemaVersion1Dot3,
    SchemaVersion1Dot4,
    SchemaVersion1Dot5,
    SchemaVersion1Dot6,
    SchemaVersion1Dot7,
)
from . import DataClassification, ExternalReference, Property, XsUri
from .bom_ref import BomRef
from .contact import OrganizationalEntity
from .dependency import Dependable
from .license import License, LicenseRepository, _LicenseRepositorySerializationHelper
from .release_note import ReleaseNotes


@serializable.serializable_class(ignore_unknown_during_deserialization=True)
class Service(Dependable):
    """
    Class that models the `service` complex type in the CycloneDX schema.

    .. note::
        See the CycloneDX schema: https://cyclonedx.org/docs/1.7/xml/#type_service
    """

    def __init__(
        self, *,
        name: str,
        bom_ref: Optional[Union[str, BomRef]] = None,
        provider: Optional[OrganizationalEntity] = None,
        group: Optional[str] = None,
        version: Optional[str] = None,
        description: Optional[str] = None,
        endpoints: Optional[Iterable[XsUri]] = None,
        authenticated: Optional[bool] = None,
        x_trust_boundary: Optional[bool] = None,
        data: Optional[Iterable[DataClassification]] = None,
        licenses: Optional[Iterable[License]] = None,
        external_references: Optional[Iterable[ExternalReference]] = None,
        properties: Optional[Iterable[Property]] = None,
        services: Optional[Iterable['Service']] = None,
        release_notes: Optional[ReleaseNotes] = None,
    ) -> None:
        self._bom_ref = _bom_ref_from_str(bom_ref)
        self.provider = provider
        self.group = group
        self.name = name
        self.version = version
        self.description = description
        self.endpoints = endpoints or []
        self.authenticated = authenticated
        self.x_trust_boundary = x_trust_boundary
        self.data = data or []
        self.licenses = licenses or []
        self.external_references = external_references or []
        self.services = services or []
        self.release_notes = release_notes
        self.properties = properties or []

    @property
    @serializable.json_name('bom-ref')
    @serializable.type_mapping(BomRef)
    @serializable.xml_attribute()
    @serializable.xml_name('bom-ref')
    def bom_ref(self) -> BomRef:
        """
        An optional identifier which can be used to reference the service elsewhere in the BOM. Uniqueness is enforced
        within all elements and children of the root-level bom element.

        Returns:
           `BomRef` unique identifier for this Service
        """
        return self._bom_ref

    @property
    @serializable.xml_sequence(1)
    def provider(self) -> Optional[OrganizationalEntity]:
        """
        Get the organization that provides the service.

        Returns:
            `OrganizationalEntity` if set else `None`
        """
        return self._provider

    @provider.setter
    def provider(self, provider: Optional[OrganizationalEntity]) -> None:
        self._provider = provider

    @property
    @serializable.xml_sequence(2)
    @serializable.xml_string(serializable.XmlStringSerializationType.NORMALIZED_STRING)
    def group(self) -> Optional[str]:
        """
        The grouping name, namespace, or identifier. This will often be a shortened, single name of the company or
        project that produced the service or domain name. Whitespace and special characters should be avoided.

        Returns:
            `str` if provided else `None`
        """
        return self._group

    @group.setter
    def group(self, group: Optional[str]) -> None:
        self._group = group

    @property
    @serializable.xml_sequence(3)
    @serializable.xml_string(serializable.XmlStringSerializationType.NORMALIZED_STRING)
    def name(self) -> str:
        """
        The name of the service. This will often be a shortened, single name of the service.

        Returns:
            `str`
        """
        return self._name

    @name.setter
    def name(self, name: str) -> None:
        self._name = name

    @property
    @serializable.xml_sequence(4)
    @serializable.xml_string(serializable.XmlStringSerializationType.NORMALIZED_STRING)
    def version(self) -> Optional[str]:
        """
        The service version.

        Returns:
            `str` if set else `None`
        """
        return self._version

    @version.setter
    def version(self, version: Optional[str]) -> None:
        self._version = version

    @property
    @serializable.xml_sequence(5)
    @serializable.xml_string(serializable.XmlStringSerializationType.NORMALIZED_STRING)
    def description(self) -> Optional[str]:
        """
        Specifies a description for the service.

        Returns:
            `str` if set else `None`
        """
        return self._description

    @description.setter
    def description(self, description: Optional[str]) -> None:
        self._description = description

    @property
    @serializable.xml_array(serializable.XmlArraySerializationType.NESTED, 'endpoint')
    @serializable.xml_sequence(6)
    def endpoints(self) -> 'SortedSet[XsUri]':
        """
        A list of endpoints URI's this service provides.

        Returns:
            Set of `XsUri`
        """
        return self._endpoints

    @endpoints.setter
    def endpoints(self, endpoints: Iterable[XsUri]) -> None:
        self._endpoints = SortedSet(endpoints)

    @property
    @serializable.xml_sequence(7)
    def authenticated(self) -> Optional[bool]:
        """
        A boolean value indicating if the service requires authentication. A value of true indicates the service
        requires authentication prior to use.

        A value of false indicates the service does not require authentication.

        Returns:
            `bool` if set else `None`
        """
        return self._authenticated

    @authenticated.setter
    def authenticated(self, authenticated: Optional[bool]) -> None:
        self._authenticated = authenticated

    @property
    @serializable.json_name('x-trust-boundary')
    @serializable.xml_name('x-trust-boundary')
    @serializable.xml_sequence(8)
    def x_trust_boundary(self) -> Optional[bool]:
        """
        A boolean value indicating if use of the service crosses a trust zone or boundary. A value of true indicates
        that by using the service, a trust boundary is crossed.

        A value of false indicates that by using the service, a trust boundary is not crossed.

        Returns:
            `bool` if set else `None`
        """
        return self._x_trust_boundary

    @x_trust_boundary.setter
    def x_trust_boundary(self, x_trust_boundary: Optional[bool]) -> None:
        self._x_trust_boundary = x_trust_boundary

    # @property
    # ...
    # @serializable.view(SchemaVersion1Dot5)
    # @serializable.xml_sequence(9)
    # def trust_zone(self) -> ...:
    #     ... # since CDX1.5
    #
    # @trust_zone.setter
    # def trust_zone(self, ...) -> None:
    #     ... # since CDX1.5

    @property
    @serializable.xml_array(serializable.XmlArraySerializationType.NESTED, 'classification')
    @serializable.xml_sequence(10)
    def data(self) -> 'SortedSet[DataClassification]':
        """
        Specifies the data classification.

        Returns:
            Set of `DataClassification`
        """
        # TODO since CDX1.5 also supports `dataflow`, not only `DataClassification`
        return self._data

    @data.setter
    def data(self, data: Iterable[DataClassification]) -> None:
        self._data = SortedSet(data)

    @property
    @serializable.type_mapping(_LicenseRepositorySerializationHelper)
    @serializable.xml_sequence(11)
    def licenses(self) -> LicenseRepository:
        """
        A optional list of statements about how this Service is licensed.

        Returns:
            Set of `LicenseChoice`
        """
        # TODO since CDX1.5 also supports `dataflow`, not only `DataClassification`
        return self._licenses

    @licenses.setter
    def licenses(self, licenses: Iterable[License]) -> None:
        self._licenses = LicenseRepository(licenses)

    @property
    @serializable.xml_array(serializable.XmlArraySerializationType.NESTED, 'reference')
    @serializable.xml_sequence(12)
    def external_references(self) -> 'SortedSet[ExternalReference]':
        """
        Provides the ability to document external references related to the Service.

        Returns:
            Set of `ExternalReference`
        """
        return self._external_references

    @external_references.setter
    def external_references(self, external_references: Iterable[ExternalReference]) -> None:
        self._external_references = SortedSet(external_references)

    @property
    @serializable.view(SchemaVersion1Dot3)
    @serializable.view(SchemaVersion1Dot4)
    @serializable.view(SchemaVersion1Dot5)
    @serializable.view(SchemaVersion1Dot6)
    @serializable.view(SchemaVersion1Dot7)
    @serializable.xml_array(serializable.XmlArraySerializationType.NESTED, 'property')
    @serializable.xml_sequence(13)
    def properties(self) -> 'SortedSet[Property]':
        """
        Provides the ability to document properties in a key/value store. This provides flexibility to include data not
        officially supported in the standard without having to use additional namespaces or create extensions.

        Return:
            Set of `Property`
        """
        return self._properties

    @properties.setter
    def properties(self, properties: Iterable[Property]) -> None:
        self._properties = SortedSet(properties)

    @property
    @serializable.xml_array(serializable.XmlArraySerializationType.NESTED, 'service')
    @serializable.xml_sequence(14)
    def services(self) -> "SortedSet['Service']":
        """
        A list of services included or deployed behind the parent service.

        This is not a dependency tree.

        It provides a way to specify a hierarchical representation of service assemblies.

        Returns:
            Set of `Service`
        """
        return self._services

    @services.setter
    def services(self, services: Iterable['Service']) -> None:
        self._services = SortedSet(services)

    @property
    @serializable.view(SchemaVersion1Dot4)
    @serializable.view(SchemaVersion1Dot5)
    @serializable.view(SchemaVersion1Dot6)
    @serializable.view(SchemaVersion1Dot7)
    @serializable.xml_sequence(15)
    def release_notes(self) -> Optional[ReleaseNotes]:
        """
        Specifies optional release notes.

        Returns:
            `ReleaseNotes` or `None`
        """
        return self._release_notes

    @release_notes.setter
    def release_notes(self, release_notes: Optional[ReleaseNotes]) -> None:
        self._release_notes = release_notes

    def __comparable_tuple(self) -> _ComparableTuple:
        return _ComparableTuple((
            self.group, self.name, self.version,
            self.bom_ref.value,
            self.provider, self.description,
            self.authenticated, _ComparableTuple(self.data), _ComparableTuple(self.endpoints),
            _ComparableTuple(self.external_references), _ComparableTuple(self.licenses),
            _ComparableTuple(self.properties), self.release_notes, _ComparableTuple(self.services),
            self.x_trust_boundary
        ))

    def __eq__(self, other: object) -> bool:
        if isinstance(other, Service):
            return self.__comparable_tuple() == other.__comparable_tuple()
        return False

    def __lt__(self, other: Any) -> bool:
        if isinstance(other, Service):
            return self.__comparable_tuple() < other.__comparable_tuple()
        return NotImplemented

    def __hash__(self) -> int:
        return hash(self.__comparable_tuple())

    def __repr__(self) -> str:
        return f'<Service bom-ref={self.bom_ref}, group={self.group}, name={self.name}, version={self.version}>'



# This file is part of CycloneDX Python Library
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# SPDX-License-Identifier: Apache-2.0
# Copyright (c) OWASP Foundation. All Rights Reserved.

from decimal import Decimal
from unittest import TestCase

from cyclonedx.exception.model import InvalidConfidenceException
from cyclonedx.model import Copyright
from cyclonedx.model.component_evidence import (
    AnalysisTechnique,
    CallStack,
    CallStackFrame,
    ComponentEvidence,
    Identity,
    IdentityField,
    Method,
    Occurrence,
)


class TestModelComponentEvidence(TestCase):

    def test_no_params(self) -> None:
        ComponentEvidence()  # Does not raise `NoPropertiesProvidedException`

    def test_identity_single(self) -> None:
        identity = Identity(field=IdentityField.NAME, confidence=Decimal('1'), concluded_value='test')
        ce = ComponentEvidence(identity=identity)
        self.assertEqual(len(ce.identity), 1)
        self.assertEqual(ce.identity.pop().field, 'name')

    def test_identity_multiple(self) -> None:
        identities = [
            Identity(field=IdentityField.NAME, confidence=Decimal('1'), concluded_value='test'),
            Identity(field=IdentityField.VERSION, confidence=Decimal('0.8'), concluded_value='1.0.0')
        ]
        ce = ComponentEvidence(identity=identities)
        self.assertEqual(len(ce.identity), 2)

    def test_identity_with_methods(self) -> None:
        """Test identity with analysis methods"""
        methods = [
            Method(
                technique=AnalysisTechnique.BINARY_ANALYSIS,  # Changed order to test sorting
                confidence=Decimal('0.9'),
                value='Found in binary'
            ),
            Method(
                technique=AnalysisTechnique.SOURCE_CODE_ANALYSIS,
                confidence=Decimal('0.8'),
                value='Found in source'
            )
        ]
        identity = Identity(field='name', confidence=Decimal('1'), methods=methods)
        self.assertEqual(len(identity.methods), 2)
        sorted_methods = sorted(methods)  # Methods should be sorted by technique name
        self.assertEqual(list(identity.methods), sorted_methods)

        # Verify first method
        method = sorted_methods[0]
        self.assertEqual(method.technique, AnalysisTechnique.BINARY_ANALYSIS)
        self.assertEqual(method.confidence, Decimal('0.9'))
        self.assertEqual(method.value, 'Found in binary')

    def test_method_sorting(self) -> None:
        """Test that methods are properly sorted by technique value"""
        methods = [
            Method(technique=AnalysisTechnique.SOURCE_CODE_ANALYSIS, confidence=Decimal('0.8')),
            Method(technique=AnalysisTechnique.BINARY_ANALYSIS, confidence=Decimal('0.9')),
            Method(technique=AnalysisTechnique.ATTESTATION, confidence=Decimal('1.0'))
        ]

        sorted_methods = sorted(methods)
        self.assertEqual(sorted_methods[0].technique, AnalysisTechnique.ATTESTATION)
        self.assertEqual(sorted_methods[1].technique, AnalysisTechnique.BINARY_ANALYSIS)
        self.assertEqual(sorted_methods[2].technique, AnalysisTechnique.SOURCE_CODE_ANALYSIS)

    def test_invalid_method_confidence(self) -> None:
        """Test that invalid confidence raises ValueError"""
        with self.assertRaises(InvalidConfidenceException):
            Method(technique=AnalysisTechnique.FILENAME, confidence=Decimal('1.5'))

    def test_occurrences(self) -> None:
        occurrence = Occurrence(location='/path/to/file', line=42)
        ce = ComponentEvidence(occurrences=[occurrence])
        self.assertEqual(len(ce.occurrences), 1)
        self.assertEqual(ce.occurrences.pop().line, 42)

    def test_callstack(self) -> None:
        frame = CallStackFrame(
            package='com.example',
            module='app',
            function='main'
        )
        stack = CallStack(frames=[frame])
        ce = ComponentEvidence(callstack=stack)
        self.assertIsNotNone(ce.callstack)
        self.assertEqual(len(ce.callstack.frames), 1)

    def test_licenses(self) -> None:
        from cyclonedx.model.license import DisjunctiveLicense
        license = DisjunctiveLicense(id='MIT')
        ce = ComponentEvidence(licenses=[license])
        self.assertEqual(len(ce.licenses), 1)

    def test_copyright(self) -> None:
        copyright = Copyright(text='(c) 2023')
        ce = ComponentEvidence(copyright=[copyright])
        self.assertEqual(len(ce.copyright), 1)
        self.assertEqual(ce.copyright.pop().text, '(c) 2023')

    def test_full_evidence(self) -> None:
        # Test with all fields populated
        identity = Identity(field=IdentityField.NAME, confidence=Decimal('1'), concluded_value='test')
        occurrence = Occurrence(location='/path/to/file', line=42)
        frame = CallStackFrame(module='app', function='main', line=1)
        stack = CallStack(frames=[frame])
        from cyclonedx.model.license import DisjunctiveLicense
        license = DisjunctiveLicense(id='MIT')
        copyright = Copyright(text='(c) 2023')

        ce = ComponentEvidence(
            identity=[identity],
            occurrences=[occurrence],
            callstack=stack,
            licenses=[license],
            copyright=[copyright]
        )

        self.assertEqual(len(ce.identity), 1)
        self.assertEqual(len(ce.occurrences), 1)
        self.assertIsNotNone(ce.callstack)
        self.assertEqual(len(ce.callstack.frames), 1)
        self.assertEqual(len(ce.licenses), 1)
        self.assertEqual(len(ce.copyright), 1)

    def test_full_evidence_with_complete_stack(self) -> None:
        identity = Identity(field=IdentityField.NAME, confidence=Decimal('1'), concluded_value='test')
        occurrence = Occurrence(location='/path/to/file', line=42)

        frame = CallStackFrame(
            package='com.example',
            module='app',
            function='main',
            parameters=['arg1', 'arg2'],
            line=1,
            column=10,
            full_filename='/path/to/file.py'
        )
        stack = CallStack(frames=[frame])

        from cyclonedx.model.license import DisjunctiveLicense
        license = DisjunctiveLicense(id='MIT')
        copyright = Copyright(text='(c) 2023')

        ce = ComponentEvidence(
            identity=[identity],
            occurrences=[occurrence],
            callstack=stack,
            licenses=[license],
            copyright=[copyright]
        )

        self.assertEqual(len(ce.identity), 1)
        self.assertEqual(len(ce.occurrences), 1)
        self.assertIsNotNone(ce.callstack)
        self.assertEqual(len(ce.callstack.frames), 1)
        self.assertEqual(ce.callstack.frames.pop().package, 'com.example')
        self.assertEqual(len(ce.licenses), 1)
        self.assertEqual(len(ce.copyright), 1)

    def test_same_1(self) -> None:
        ce_1 = ComponentEvidence(copyright=[Copyright(text='Commercial')])
        ce_2 = ComponentEvidence(copyright=[Copyright(text='Commercial')])
        self.assertEqual(hash(ce_1), hash(ce_2))
        self.assertTrue(ce_1 == ce_2)

    def test_same_2(self) -> None:
        ce_1 = ComponentEvidence(copyright=[Copyright(text='Commercial'), Copyright(text='Commercial 2')])
        ce_2 = ComponentEvidence(copyright=[Copyright(text='Commercial 2'), Copyright(text='Commercial')])
        self.assertEqual(hash(ce_1), hash(ce_2))
        self.assertTrue(ce_1 == ce_2)

    def test_not_same_1(self) -> None:
        ce_1 = ComponentEvidence(copyright=[Copyright(text='Commercial')])
        ce_2 = ComponentEvidence(copyright=[Copyright(text='Commercial 2')])
        self.assertNotEqual(hash(ce_1), hash(ce_2))
        self.assertFalse(ce_1 == ce_2)

    def test_component_evidence_sorting(self) -> None:
        """Test that ComponentEvidence instances can be sorted without triggering TypeError"""
        ce_1 = ComponentEvidence(copyright=[Copyright(text='Copyright A')])
        ce_2 = ComponentEvidence(copyright=[Copyright(text='Copyright B')])
        ce_3 = ComponentEvidence(copyright=[Copyright(text='Copyright C')])

        # This should not raise TypeError: '<' not supported between instances
        evidence_list = [ce_3, ce_1, ce_2]
        sorted_evidence = sorted(evidence_list)
        self.assertEqual(len(sorted_evidence), 3)


class TestModelCallStackFrame(TestCase):

    def test_fields(self) -> None:
        # Test CallStackFrame with required fields
        frame = CallStackFrame(
            package='com.example',
            module='app',
            function='main',
            parameters=['arg1', 'arg2'],
            line=1,
            column=10,
            full_filename='/path/to/file.py'
        )
        self.assertEqual(frame.package, 'com.example')
        self.assertEqual(frame.module, 'app')
        self.assertEqual(frame.function, 'main')
        self.assertEqual(len(frame.parameters), 2)
        self.assertEqual(frame.line, 1)
        self.assertEqual(frame.column, 10)
        self.assertEqual(frame.full_filename, '/path/to/file.py')

    def test_module_required(self) -> None:
        """Test that module is the only required field"""
        frame = CallStackFrame(module='app')  # Only mandatory field
        self.assertEqual(frame.module, 'app')
        self.assertIsNone(frame.package)
        self.assertIsNone(frame.function)
        self.assertEqual(len(frame.parameters), 0)
        self.assertIsNone(frame.line)
        self.assertIsNone(frame.column)
        self.assertIsNone(frame.full_filename)

    def test_callstack_frame_sorting(self) -> None:
        """Test that CallStackFrame instances can be sorted without triggering TypeError"""
        frame1 = CallStackFrame(module='app_a', function='func_a')
        frame2 = CallStackFrame(module='app_b', function='func_b')
        frame3 = CallStackFrame(module='app_c', function='func_c')

        # This should not raise TypeError: '<' not supported between instances
        frame_list = [frame3, frame1, frame2]
        sorted_frames = sorted(frame_list)
        self.assertEqual(len(sorted_frames), 3)
